#pragma once
#include <sstream>
#include <thread>
#include <mutex>
#include <condition_variable>

namespace fcf {

enum {
  LOG_LEVEL_TST,
  LOG_LEVEL_ERR,
  LOG_LEVEL_WRN,
  LOG_LEVEL_LOG,
  LOG_LEVEL_TRC
};


namespace details {

class log_t {
public:
  class log_buffer_t {
  public:
    inline log_buffer_t(log_t& a_log, int a_level) 
      : _log(a_log), _level(a_level), _skeep(false) {
    }

    inline log_buffer_t(log_buffer_t& a_log_buffer)
      : _log(a_log_buffer._log), _level(a_log_buffer._level), _skeep(false) {
      a_log_buffer._skeep = true;
      _ss << a_log_buffer._ss.str();
    }

    template <typename Ty>
    log_buffer_t& operator << (const Ty& a_value){
      if (_level > _log.get_level())
        return *this;
      _ss << a_value;
      return *this;
    }

    inline ~log_buffer_t(){
      if (_level > _log.get_level())
        return;
      if (_skeep)
        return;
      _log.write(_level, _ss.str());
    }

  private:
    int               _life_time_days;
    log_t&            _log;
    const int         _level;
    std::stringstream _ss;
    bool              _skeep;
  };

  log_t();
  ~log_t();
  void        write(int a_level, const std::string& a_str);
  void        set_level(int a_level);
  int         get_level();
  void        set_file_prefix(const std::string& a_file_prefix);
  std::string get_file_preffix();
  void        set_life_time(int a_life_time);
  int         get_life_time();  
  void        remove_old_files();

private:
  void        _remove_old_files(std::unique_lock<std::mutex>& a_lock);
  void        _gc();
  
  bool                    _stop;
  int                     _level;
  int                     _life_time;
  std::string             _file_prefix;
  std::mutex              _mutex;
  std::condition_variable _condition;
  std::thread             _gc_thread;  


};

extern log_t logger;


class log_input_t {
public:
  inline log_input_t(log_t& a_log, int a_level)
    : _log(a_log), _level(a_level){ 
  }

  inline log_t& get_logger() {
    return _log;
  }

  template <typename Ty>
  log_t::log_buffer_t operator <<(const Ty& a_value){
    log_t::log_buffer_t lb(_log, _level);
    lb << a_value;
    return lb;
  }

private:
  log_t&  _log;
  int     _level;
};

}


extern details::log_input_t tst;
extern details::log_input_t err;
extern details::log_input_t wrn;
extern details::log_input_t log;
extern details::log_input_t trc;

}
